#!/usr/bin/env python
# coding: utf-8


import os
import pickle
import pandas as pd
import numpy as np
import scipy

from arboreto.utils import load_tf_names
from arboreto.algo import grnboost2

from pyscenic.rnkdb import FeatherRankingDatabase as RankingDatabase
from pyscenic.utils import modules_from_adjacencies, load_motifs
from pyscenic.prune import prune2df, df2regulons
from pyscenic.aucell import aucell

# !!!!
data_to_use="all_integrated"
save_folder="all_integrated"
# !!!!

RESOURCES_FOLDER="/home/jupyter/scenic/resources"
MOTIF_ANNOTATIONS_FNAME = os.path.join(RESOURCES_FOLDER, "motifs-v9-nr.hgnc-m0.001-o0.0.tbl")
TFS_FNAME = os.path.join(RESOURCES_FOLDER, 'hs_hgnc_curated_tfs.txt')
TF_LENGTH = sum(1 for line in open(TFS_FNAME))*2
EXP_FNAME = os.path.join("/home/jupyter/scenic/",data_to_use+"_expr_mat.csv")
dbs = [RankingDatabase(os.path.join(RESOURCES_FOLDER,"hg38__refseq-r80__500bp_up_and_100bp_down_tss.mc9nr.feather"), "hg38__refseq-r80__500bp_up_and_100bp_down_tss.mc9nr.feather"), RankingDatabase(os.path.join(RESOURCES_FOLDER, "hg38__refseq-r80__10kb_up_and_down_tss.mc9nr.feather"), "hg38__refseq-r80__10kb_up_and_down_tss.mc9nr.feather")]

num_iterations = 10

### create iterations of adjacency matrices
for i in range(num_iterations):
    DATA_FOLDER="/home/jupyter/scenic/"+save_folder+"/iterations/"+str(i)
    ADJACENCIES_FNAME = os.path.join(DATA_FOLDER, "adjacencies.pickle")
    
    if not os.path.exists(str(DATA_FOLDER)):
        os.makedirs(str(DATA_FOLDER))
        
    #read matrix
    ex_matrix = pd.read_csv(EXP_FNAME, sep=',', header=0, index_col=0).T

    tf_names = load_tf_names(TFS_FNAME)

    adjacencies = grnboost2(ex_matrix, tf_names=tf_names, verbose=True)
    
    with open(ADJACENCIES_FNAME, 'wb') as handle:
        pickle.dump(adjacencies, handle, protocol=pickle.HIGHEST_PROTOCOL)

### intersect produced adjacency matrices --- find common TF-target links       
common_tf_gene = None
for i in range(num_iterations):
    with open("/home/jupyter/scenic/"+save_folder+"/iterations/"+str(i)+"/adjacencies.pickle", 'rb') as handle:
        temp_adj = pickle.load(handle)
    if(common_tf_gene is None):
        common_tf_gene = set(temp_adj.TF+"*"+temp_adj.target)
    else:
        common_tf_gene = common_tf_gene.intersection(set(temp_adj.TF+"*"+temp_adj.target))
print("Done with intersection")

### calculate median of common TF-target importance
arr = None
for i in range(num_iterations):
    with open("/home/jupyter/scenic/"+save_folder+"/iterations/"+str(i)+"/adjacencies.pickle", 'rb') as handle:
        temp_adj = pickle.load(handle)
    idx_list = [True if x in common_tf_gene else False for x in (temp_adj.TF+"*"+temp_adj.target)]
    temp_adj = temp_adj[idx_list]
    temp_adj = temp_adj.sort_values(["TF","target"])
    if(arr is None):
        arr = np.empty((len(temp_adj.index),num_iterations))
        tf_order = temp_adj.TF
        gene_order = temp_adj.target
    assert all(temp_adj.TF == tf_order)
    assert all(temp_adj.target == gene_order)
    arr[:,i] = temp_adj.importance
med_arr = np.median(arr,1)
new_adj = pd.DataFrame(data = {"TF": tf_order, "target": gene_order, "importance":med_arr})
new_adj = new_adj.sort_values(["importance"], ascending=False)

with open("/home/jupyter/scenic/"+save_folder+"/adjacencies.pickle", 'wb') as handle:
    pickle.dump(new_adj, handle, protocol=pickle.HIGHEST_PROTOCOL)
    
    
### run standard scenic analysis with meta-adjancencies
DATA_FOLDER="/home/jupyter/scenic/"+save_folder
ADJACENCIES_FNAME = os.path.join(DATA_FOLDER, "adjacencies.pickle")
MODULES_FNAME = os.path.join(DATA_FOLDER, "modules.pickle")
REGULONS_FNAME = os.path.join(DATA_FOLDER, "regulons.pickle")
MOTIFS_FNAME = os.path.join(DATA_FOLDER, "motifs.csv")
DF_FNAME = os.path.join(DATA_FOLDER, "df.pickle")
AUCELL_MTX_FNAME = os.path.join(DATA_FOLDER, "auc_mtx.pickle")
AUCELL_FNAME = os.path.join(DATA_FOLDER, "auc_mtx.csv")


if not os.path.exists(str(DATA_FOLDER)):
    os.makedirs(str(DATA_FOLDER))

#read matrix
ex_matrix = pd.read_csv(EXP_FNAME, sep=',', header=0, index_col=0).T
tf_names = load_tf_names(TFS_FNAME)

#Step1
with open(ADJACENCIES_FNAME, 'rb') as handle:
    adjacencies = pickle.load(handle)

modules = list(modules_from_adjacencies(adjacencies, ex_matrix, rho_mask_dropouts=True, keep_only_activating=False))
with open(MODULES_FNAME, 'wb') as handle:
    pickle.dump(modules, handle, protocol=pickle.HIGHEST_PROTOCOL)

#Step2
df = prune2df(rnkdbs=dbs, modules=modules, motif_annotations_fname=MOTIF_ANNOTATIONS_FNAME, client_or_address="custom_multiprocessing", num_workers=30)
with open(DF_FNAME, 'wb') as handle:
    pickle.dump(df, handle, protocol=pickle.HIGHEST_PROTOCOL)
df.to_csv(MOTIFS_FNAME)

regulons = df2regulons(df)
regu_tf = [itm.name for itm in regulons]
regu_genes = [itm.genes for itm in regulons]
regu_df = pd.DataFrame(list(zip(regu_tf, regu_genes)), columns=["TF", "Genes"])
regu_df.to_csv(os.path.join(DATA_FOLDER,"regulons_genes.csv"), index=False)

with open(REGULONS_FNAME, 'wb') as handle:
    pickle.dump(regulons, handle, protocol=pickle.HIGHEST_PROTOCOL)

#Step3
auc_mtx = aucell(ex_matrix, regulons, num_workers=20)
with open(AUCELL_MTX_FNAME, 'wb') as handle:
    pickle.dump(auc_mtx, handle, protocol=pickle.HIGHEST_PROTOCOL)

auc_mtx.to_csv(AUCELL_FNAME)

